// ----------------------------------------------------------------------------
//   LCD functions
// ----------------------------------------------------------------------------
#include "Display.h"

Tlcd_glob lcdGlob;


/*
     PortD: 8xData
     PB0  LCD RS
     PB2  LCD enable Z12 (upper)
     PB3  LCD enable Z34 (lower)

     PortD: 8xData
     PC2  LCD RS
     PC0  LCD enable Z12 (upper)
     PC1  LCD enable Z34 (lower)
*/


#define  LCD_ENABLE    PORTC = ((PORTC & ~(BIT0|BIT1)) | lcdGlob.enablePattern)
#define  LCD_DISABLE   {PORTC &= ~BIT0; PORTC &= ~BIT1;} 
#define  LCD_DATA      PORTD
#define  LCD_RS_ON     PORTC |=  BIT2
#define  LCD_RS_OFF    PORTC &= ~BIT2
//#define  LCD_WRITE     PORTB &= ~BIT4
//#define  LCD_READ      PORTB |=  BIT4

//in .h Datei anpassen:
//#define  LCD_UPPER     lcdGlob.enablePattern = BIT0
//#define  LCD_LOWER     lcdGlob.enablePattern = BIT1


// ---------------------------------------------------------------------------
/*! Send byte to display.
    Depending on the value of LCD_RS the byte is written to display(LCD_RS=1) or register (LCD_RS=0)

   \param b is the byte to be send.
   \note the BUSY-Flag isn't ckecked (since we can't read from display with R/W hard-wired to GND)
*/// -------------------------------------------------------------------------
void lcdWrite(unsigned char b) {
//unsigned char i;
  
      asm volatile ("nop"); // 1 NOP bei 8MHz braucht etwa  125ns.  t_AS: 40ns min  #BATRON, page 57
      asm volatile ("nop");
      LCD_ENABLE;

      LCD_DATA = b;
      asm volatile ("nop");
      asm volatile ("nop");
      asm volatile ("nop");
      asm volatile ("nop");  //t_DSW: 80ns min.
      _delay_us(4);
      LCD_DISABLE;
      //t_H: 10ns min
      //t_CYC 500ns min

      //Delay >37s  
      _delay_us(50);
      
      // CAUTION: we do not check the BUSY-Flag (since we can't read from display)
      // CAUTION: so for HOME and CLEAR commands the user has to add 2ms delay!
}



// ---------------------------------------------------------------------------
void cmd2Lcd(unsigned char cmd) {
  LCD_RS_OFF;
  _delay_us(80);
  lcdWrite(cmd);
  LCD_RS_ON;
  _delay_us(2050);
}
// ---------------------------------------------------------------------------
/*! Clear LCD and set cursor home
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------

void clearLcd(void) {
  cmd2Lcd(0x01);
}

// ---------------------------------------------------------------------------
/*! Clear upper and lower LCD and set cursor home on upper
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------
void clearFullLcd(void) {
  LCD_LOWER;
  cmd2Lcd(0x01);
  LCD_UPPER;
  cmd2Lcd(0x01);
}


// ---------------------------------------------------------------------------
/*! Set cursor home
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------
void home1Lcd(void) {
  cmd2Lcd(0x02);
}
// ---------------------------------------------------------------------------
/*! Set cursor at start of Line 2
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------
void home2Lcd(void) {
  cmd2Lcd(192);
}


		

// ---------------------------------------------------------------------------
/*! Initialize display (2 Lines, 8 Bit, no shift, increment)
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------
void lcdInitHalf(void) {

  LCD_RS_OFF;
  _delay_us(300);
  LCD_ENABLE;

  //Delay >15ms
  _delay_ms(30);
  
  lcdWrite(0x30); // DB5=1 DB4=1        # see BATRON-Specification for HD44780, page 58
  _delay_ms(15);    // wait more than 4.1ms

  lcdWrite(0x30); // DB5=1 DB4=1
  _delay_ms(15);    // wait more than 100us

  lcdWrite(0x30);
  _delay_ms(15);	   //just to be sure...

  lcdWrite(0x38); //2 Lines, 8Bit Data,  
  _delay_us(1000);    //just for paranoia
  lcdWrite(0x38); //2 Lines, 8Bit Data,  
  _delay_us(1000);    //just for paranoia

  lcdWrite(0x0c); // Display on, Cursor off
  _delay_us(1000);    //just for paranoia  
/*
  lcdWrite(0x01); // Display clear
  _delay_ms(3);
*/
  lcdWrite(0x06); // Entry mode; no shift
  _delay_us(2000);
  
  LCD_RS_ON;
}
// ---------------------------------------------------------------------------
void lcdInit(void) {
//    LCD_WRITE;
    LCD_UPPER;
    lcdInitHalf();
    LCD_LOWER;
    lcdInitHalf();
    LCD_UPPER;
}
//----- String senden ----------------------------------------------------------
void sendStringToLCD(char *str)
{
  unsigned char i;

  //Zeichen fuer Zeichen senden (ohne abschliessende '\0')
  i = 0;
  while (str[i] !=0)
  {
    lcdWrite(str[i]);
    i++;
    // mehr als 27 Zeichen werden nie an das Display gesendet
    if (i > 27){
      break;
    }
  }
}





//----- Schreiben aufs Display  --------------------------------------------------
void sendMessageToLCD(void)
{
// Setzen von lcdGlob.showMessage auf 1 startet das Schreiben des Textes 
// aus lcdGlob.message. Wenn fertig, wird lcdGlob.showMessage auf 0 gesetzt.

/*Spezialzeichen im String:
  o  = Oberer Display-Controller (Zeile1+2)
  u  = Unterer Display-Controller (Zeile 3+4)
  *  = ClearLCD (auf gerade eingestelltem Controller)
  1  = Home 1. Zeile 
  2  = Home 2. Zeile
  3  = Home 3. Zeile
  4  = Home 4. Zeile
  Gyx= GotoYX x:1..4  y:@..Z
  C  = Cursor Ein (auf gerade eingestelltem Controller) 
  B  = Cursor Ein + Blinken(auf gerade eingestelltem Controller) 
  c  = Cursor Aus (auf gerade eingestelltem Controller) 
  Die HOME-Befehle schalten auch die Displayhlfte um
*/

	
static unsigned char i;
static unsigned char cmd;
static unsigned char SendStatus=0;
unsigned char c;	
unsigned char offset;

  
  switch (SendStatus){
    case 0: 
        i=0; 
        if (1==lcdGlob.showMessage){
          lcdGlob.showMessage=2;
          SendStatus=1;
        }else
        if (2==lcdGlob.showMessage){
          lcdGlob.showMessage=0;
        }
        break;
    case 1: //Zeichen fuer Zeichen senden (ohne abschliessende '\0')
      c=lcdGlob.message[i++];
      if (!c){ //Textende
        SendStatus=0;
        break;
      }
      if (''==c){ //Steuerkommando
        c=lcdGlob.message[i++];
     	switch(c){
          case 'G':  
            c=lcdGlob.message[i++]; //Zeile (Y)
            switch(c){
              default: LCD_UPPER; offset=0;  break;		
              case '2':LCD_UPPER; offset=64; break;		
              case '3':LCD_LOWER; offset=0;  break;		
              case '4':LCD_LOWER; offset=64; break;		
            }
            c=lcdGlob.message[i++]-64; //Spalte (X)
	        cmd=c + offset +0x80; //GotoYX 	      LCD_SET_DDRAM + x + (y * 64));	
	        SendStatus=2;
	        break;
          case 'C': //Cursor EIN
  	        cmd=0x0E;
	        SendStatus=4;
            break;		
          case 'B': //Cursor EIN + Blinken
  	        cmd=0x0F;
	        SendStatus=4;
            break;		
          case 'c': //Cursor AUS
  	        cmd=0x0C;
	        SendStatus=4;
            break;		
          case 'o': LCD_UPPER; break;		
          case 'u': LCD_LOWER; break;		
          case '*': 
	        cmd=0x01; //CLEAR DISPLAY
	        SendStatus=2;
	        break;
	      case '1':
	        LCD_UPPER;
	        cmd=2; //Home 1
   	        SendStatus=2;
	        break;
    	  case '2':
    	    LCD_UPPER;
    	    cmd=192; //Home 2
    	    SendStatus=2;
    	    break;
    	  case '3':
    	    LCD_LOWER;
    	    cmd=2; //Home 1
    	    SendStatus=2;
    	    break;
    	  case '4':
    	    LCD_LOWER;
    	    cmd=192; //Home 2
    	    SendStatus=2;
    	    break;
       	}		
      }else{ //Normales Zeichen
    	// ggf. Sonderzeichen umsetzen
        if (c >= 0xB0){
        	if (c == 0xE4) c=0xE1;		// 
        	if (c == 0xF6) c=0xEF;		// 
        	if (c == 0xFC) c=0xF5;		// 
        	if (c == 0xDF) c=0xE2;		// 
        	if (c == 0xB0) c=0xDF;		//  (Grad)
        }
        lcdWrite(c);
      }
      break;
      
    case 2: //CMD zum Display   
      LCD_RS_OFF;
      lcdWrite(cmd);
      LCD_RS_ON;
      SendStatus=3;
      break;

    case 3: //Warte 2-3ms, bis CMD ausgefuehrt    
      _delay_ms(2);
	    SendStatus=1;
      break;

    case 4: //Fast CMD zum Display   
      LCD_RS_OFF;
      lcdWrite(cmd);
      LCD_RS_ON;
      SendStatus=1;
      break;
  }	  
}

//EOF
