/* Tacho mit Jupiter-GPS-Chip aus Radiosonde

 Realisiert mit ATmega88

 Version 1.0    2013-04-30
 
 
 
 Interner RC Oszillator mit 1MHz  (Teiler/8 eingeschaltet)

                     (Reset) PC6 1  28 PC5                    ADC5
                         RX  PD0 2  27 PC4                    ADC4
                         TX  PD1 3  26 PC3                    ADC3
                  LCD_Enable PD2 4  25 PC2                    ADC2
                  LCD_RS     PD3 5  24 PC1                    ADC1
                  LCD_Data4  PD4 6  23 PC0                    ADC0
                             VCC 7  22 GND
                             GND 8  21 AREF
                             PB6 9  20 AVCC
                             PB7 10 19 PB5   (SCK)       
                   LCD_Data5 PD5 11 18 PB4   (MISO)       
                   LCD_Data6 PD6 12 17 PB3   (MOSI)      
                   LCD_Data7 PD7 13 16 PB2       
                Every second PB0 14 15 PB1   Licht                                             

*/




#define F_CPU 1000000UL    /* CPU clock in Hertz */

#include <util/delay.h> 
#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h> //memset
//#include <avr/pgmspace.h>

FUSES = {0x62, 0xDF, 0xFF}; //Lo Hi Ext

#define RX_BUFFERSIZE 82 // Jeder Satz kann inklusive des fhrenden "$" und den beiden CR/LF bis zu 82 Zeichen enthalten.
#define  LCD_ENABLE    {PORTD |= BIT2;}
#define  LCD_DISABLE   {PORTD &= ~BIT2;} 
#define  LCD_DATA      PORTD
#define  LCD_RS_ON     PORTD |=  BIT3
#define  LCD_RS_OFF    PORTD &= ~BIT3

#define LICHT_AUS PORTB|=BIT1
#define LICHT_AN  PORTB&=~BIT1


#define BIT0 1
#define BIT1 2
#define BIT2 4
#define BIT3 8
#define BIT4 16
#define BIT5 32
#define BIT6 64
#define BIT7 128

#define U8 unsigned char
#define I8   signed char
#define U16 unsigned int
#define I16   signed int
#define I32   signed long 
#define U32 unsigned long 

typedef struct{
  U8 RxBuffer[RX_BUFFERSIZE];
  U8 RxBufferRead;
  U8 RxBufferWrite;
  U8 bToggle1000ms;
  I16 iTimeTic1ms;
}T_global;


volatile T_global  g;
U8 txt[7];  //Universell zum Senden genutzter Buffer

//-----------------------------------------------------------------------------
void initHW(void){
//-----------------------------------------------------------------------------

  //Globale var auf 0
  memset((U8 *)(&g), 0, sizeof(T_global));  

              
  DDRB = BIT1;
  DDRC = 0;
  DDRD = BIT1 |BIT2 | BIT3 | BIT4 | BIT5 | BIT6 | BIT7;

  //Pull-up Widerstnde einschalten 
  PORTB = 0xff; //Alle (incl. Licht)
  PORTC = BIT0 | BIT1 | BIT2 | BIT3 | BIT4 | BIT5;
  PORTD = 0;  //Keine

  
//  PCICR |=(1<<PCIE0);  // enables range PCINT0..7        (Drehgeber)
//  PCMSK0=0xff;         // enables each pin of PCINT0..7
  
  EIMSK = 0;  //Keine externen Interrupts
  
  //Timer0: 1KHz Takt
  TCCR0A = (1<<WGM01); //CTC Mode(2),
  TCCR0B = (1<<CS01) | (1<<CS00); // clk/64  (250KHz bei 16MHz Takt)
  OCR0A = 249;
  TIMSK0 = (1<<OCIE0A);

  //Timer1: 10Hz Takt
  TCCR1A = 0;
  TCCR1B = (1<<WGM12) | (1<<CS11)| (1<<CS10); //CTC Mode(4), clk/64  (250KHz bei 16MHz Takt)
  OCR1A = 24999;
  TIMSK1 = (1<<OCIE1A);

}




/*
Die Prfsumme besteht aus einem "*" und eine Hexadezimalzahl, die sich durch ein (bitweise) 
exklusiv-oder aller Zeichen zwischen dem "$" und dem "*" berechnen. 

*/
#if 0
// ---------------------------------------------------------------------------
U8 nmea_checksum(void){
// ---------------------------------------------------------------------------

  // char * st1 = "GPZDA,201530.00,04,07,2002,00,00";
  //char * st1 ="GPGGA,235957.000,0000.0000,N,00000.0000,E,0,00,99.0,0082.0,M,18.0,M,,,";
  char * st1 ="PSRF103,04,00,01,01";
  int i;
  unsigned char XOR;
  unsigned long ilen;

  ilen = strlen(st1);

  for (XOR=0,i=0;i<ilen;i++)
  {
    XOR ^= st1[i];
  }

  return(XOR);
}
#endif





// ---------------------------------------------------------------------------
/*! Send nibble to display.
    Depending on the value of LCD_RS the byte is written to display(LCD_RS=1) or register (LCD_RS=0)

   \param b is the nibble to be send. Only bits 4..7 are used
*/// -------------------------------------------------------------------------
void lcdWriteNibble(U8 b) { //b: only Bits 4..7 are used
      asm volatile ("nop"); // 1 NOP bei 8MHz braucht etwa  125ns.  t_AS: 40ns min 
      LCD_ENABLE;
      LCD_DATA = (LCD_DATA & 0x0F) | (b & 0xf0); 
      _delay_us(4); //t_DSW: 80ns min.
      LCD_DISABLE;
      //t_H: 10ns min
      //t_CYC 500ns min
      _delay_us(2);
}

/*! Send byte to display.
    Depending on the value of LCD_RS the byte is written to display(LCD_RS=1) or register (LCD_RS=0)

   \param b is the byte to be send.
   \note the BUSY-Flag isn't checked (since we can't read from display with R/W hard-wired to GND)
*/// -------------------------------------------------------------------------
void lcdWrite(U8 b) { 
  lcdWriteNibble(b);
  lcdWriteNibble(b<<4);
  //Delay >37s  
  _delay_us(50);
  // CAUTION: we do not check the BUSY-Flag (since we can't read from display)
  // CAUTION: so for HOME and CLEAR commands the user has to add 2ms delay!
}

// ---------------------------------------------------------------------------
void lcdWrite9(U8 n){ //Write one digit. Allowed are values from 0..9
  lcdWrite(48+n);
}

// ---------------------------------------------------------------------------
void cmd2Lcd(U8 cmd) {
  LCD_RS_OFF;
  _delay_us(80);
  lcdWrite(cmd);
  LCD_RS_ON;
  _delay_us(2050);
}
// ---------------------------------------------------------------------------
/*! Clear LCD and set cursor home
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------

void clearLcd(void) {
  cmd2Lcd(0x01);
}


// ---------------------------------------------------------------------------
/*! Set cursor home in line 1
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------
void home1Lcd(void) {
  cmd2Lcd(2);
}
// ---------------------------------------------------------------------------
/*! Set cursor at start of Line 2
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------
void home2Lcd(void) {
  cmd2Lcd(192);
}


// ---------------------------------------------------------------------------
/*! Initialize display (2 Lines, 4 Bit, no shift, increment)
   (HD 44780 based displays only)
*/// -------------------------------------------------------------------------
void lcdInit(void) {

  LCD_RS_OFF;
  _delay_us(300);
  LCD_ENABLE;

  //Delay >15ms - passierte eigentlich bereits in cldInit
  _delay_ms(20);
  
  lcdWriteNibble(0x30); // DB5=1 DB4=1        # see BATRON-Specification for HD44780, page 58
  _delay_ms(15);    // wait more than 4.1ms

  lcdWriteNibble(0x30); // DB5=1 DB4=1
  _delay_ms(15);    // wait more than 100us

  lcdWriteNibble(0x30); // DB5=1 DB4=1
  _delay_ms(15);    // wait more than 100us


  lcdWrite(0x28); //2 Lines, 4Bit Data,   0 0 1 DL  N F * *  DL:0=4Bit, 1=8Bit  n:0=1Zeile, 1:2Zeilen  F:0=5x7, 1=5x11 Pixel
  _delay_us(1000);    //just for paranoia
  lcdWrite(0x28); //2 Lines, 4Bit Data,  
  _delay_us(1000);    //just for paranoia

  lcdWrite(0x08); // Display off
  _delay_us(1000);    //just for paranoia  

  lcdWrite(0x0c); // Display on, Cursor + Blink off
  _delay_us(1000);    //just for paranoia  
/*
  lcdWrite(0x01); // Display clear
  _delay_ms(3);
*/
  lcdWrite(0x06); // Entry mode; no shift
  _delay_us(2000);
  
  LCD_RS_ON;
}

//----- String sofort senden ----------------------------------------------------------
void sendStringToLCD(U8 *str)
{
  U8 i;

  //Zeichen fuer Zeichen senden (ohne abschliessende '\0')
  i = 0;
  while (str[i] !=0)
  {
    lcdWrite(str[i]);
    i++;
    // mehr als 27 Zeichen werden nie an das Display gesendet
    if (i > 27){
      break;
    }
  }
}


//-----------------------------------------------------------------------------
ISR(TIMER1_COMPA_vect){ //100ms Timer 
//-----------------------------------------------------------------------------
static U8 t10=10;
  t10--;
  if (0==t10){
    t10 = 10;
    g.bToggle1000ms = !g.bToggle1000ms;
  }
}
//-----------------------------------------------------------------------------
ISR(TIMER0_COMPA_vect){ //Millisekunden Timer 
//-----------------------------------------------------------------------------
  g.iTimeTic1ms++;
  if (31000 == g.iTimeTic1ms) g.iTimeTic1ms=0;
}




//--------------------------------------------------------
ISR(USART_RX_vect){ //Zeichen von RS232 empfangen
//--------------------------------------------------------
  g.RxBuffer[g.RxBufferWrite++] = UDR0;
  if (RX_BUFFERSIZE == g.RxBufferWrite){
    g.RxBufferWrite=0;
  }
}

/*
//--------------------------------------------------------
U8 RxDataAvail(void){ //Wieviele Zeichen sind im Puffer
//--------------------------------------------------------
  return (g.RxBufferWrite + RX_BUFFERSIZE - g.RxBufferRead) % RX_BUFFERSIZE;
}
*/

/**************************************************************************/
/* Zeitdifferenz mit 1ms Granularitaet */
/* Laufzeit bis einschl. 30s zulaessig */
/**************************************************************************/
I16 deltaT_ms(I16 startZeit) {
  I16 delta;
  cli();
  delta = g.iTimeTic1ms;
  sei();
  delta -= startZeit;
  if (delta < 0) delta += 31000;
  return delta;
}
//--------------------------------------------------------
void USART_Init(void){
//--------------------------------------------------------
 
  /* Set baud rate: round(f_osc / (8*baudrate)) - 1  wenn U2X=1*/

  UBRR0H = 0;
  UBRR0L = 12; //9600 Baud bei 1MHz

  UCSR0A=(1<<U2X0); //set U2X

  /* Set frame format:          1stop bit   8data,*/
  UCSR0C =  (3<<UCSZ00);

  // Tx+RX enable. RX interrupt
  UCSR0B = (1<<TXEN0)|(1<<RXEN0)|(1<<RXCIE0);
}

//--------------------------------------------------------
void USART_Transmit(U8 data ){
//--------------------------------------------------------
  /* Wait for empty transmit buffer */
  while ( !( UCSR0A & (1<<UDRE0)) )
  ;
  /* Put data into buffer, sends the data */
  UDR0 = data;
}


//-----------------------------------------------------------------------------
void sendTxt(U8 *txt){
//-----------------------------------------------------------------------------
  while (*txt){
    USART_Transmit(*txt++);
  }
}


//-----------------------------------------------------------------------------
void sendNMEAcommand(U8 *txt){
//-----------------------------------------------------------------------------
U8 xor=0;
U8 c=0;
U8 x;
  USART_Transmit('$');

  while (*txt){
    c = *txt++;
    USART_Transmit(c);
    xor ^= c;
  }
  USART_Transmit('*');
  //Checksumme als Hex
  x=c/16;
  if (x>9) x+=55; else x+=48;
  USART_Transmit(x);
  x=c & 15;
  if (x>9) x+=55; else x+=48;
  USART_Transmit(x);
  USART_Transmit(13);
  USART_Transmit(10);
}

//-----------------------------------------------------------------------------
void setupNMEA(void){
//-----------------------------------------------------------------------------

#if 1
  USART_Transmit(0xA0);
  USART_Transmit(0xA2);

  USART_Transmit(0);
  USART_Transmit(24);   //Payload length

  USART_Transmit(0x81); //Message ID
  USART_Transmit(0x00); //Mode 0
  USART_Transmit(0x00); //GGA
  USART_Transmit(0x00); // " Checksum 
  USART_Transmit(0x00); //GLL
  USART_Transmit(0x00); // " Checksum 
  USART_Transmit(0x00); //GSA
  USART_Transmit(0x00); // " Checksum 
  USART_Transmit(0x00); //GSV
  USART_Transmit(0x00); // " Checksum 
  USART_Transmit(0x01); //RMC
  USART_Transmit(0x01); // " Checksum 
  USART_Transmit(0x01); //VTG
  USART_Transmit(0x01); // " Checksum 
  USART_Transmit(0x00); //MSS
  USART_Transmit(0x00); // " Checksum 
  USART_Transmit(0x00); //EPE
  USART_Transmit(0x00); // " Checksum 
  USART_Transmit(0x00); //ZDA
  USART_Transmit(0x00); // " Checksum 
  USART_Transmit(0x00); //unused
  USART_Transmit(0x00); //unused
  USART_Transmit(37 ); //9600 / 256  Bitrate 9600 
  USART_Transmit(128); //9600 & 0xff);


  USART_Transmit(1);
  USART_Transmit(42);  //Checksumme

  USART_Transmit(0xb0);
  USART_Transmit(0xb3);
#else
  sendNMEAcommand((U8*)"PSRF100,1,9600,8,1,0");//NMEA, 9600/8N1
  _delay_ms(10);
  //sendTxt("$PSRF105,1*3E");//Debug ON
  //_delay_ms(10);
  sendNMEAcommand((U8*)"PSRF103,05,00,01,01");  //enable VTG
  _delay_ms(10);
  sendNMEAcommand((U8*)"PSRF103,04,00,01,01");  //enable RMC 
  _delay_ms(10);
#endif  

}
//-----------------------------------------------------------------------------
U8 getNextChar(void){
//-----------------------------------------------------------------------------
U8 c;
  c = g.RxBuffer[g.RxBufferRead++]; 
  if (RX_BUFFERSIZE == g.RxBufferRead){
    g.RxBufferRead = 0;
  }
  return c;
}
//-----------------------------------------------------------------------------
int main(void){
//-----------------------------------------------------------------------------
static U8 status=0;
U8 c;
U8 n=0;
float value=0;
float mult=0.1;
float kmh=0;
U16 x;
U8 hh=9;
U8 mm=7;
U8 ss=5;
U8 h=8;
U8 m=6;
U8 s=4;
U8 z=3;    
U8 invalidNMEA=3;
U8 s1=0;
  cli();
  initHW();
  USART_Init();
  lcdInit();

  clearLcd();
  sendStringToLCD((U8 *)"  Zabex GPS Tacho");
                 //123456789 1234567890
  sei();

  while(1){
    if (g.bToggle1000ms != s1){
      s1 = g.bToggle1000ms;
      if (invalidNMEA){
        setupNMEA();
      //}else{
        // Tx disable
       // UCSR0B = (1<<RXEN0)|(1<<RXCIE0);
       // PORTD &=~BIT1; //TX auf Low.(Strom sparen)
      }
    }

    //Empfangene Zeichen verarbeiten
    if (g.RxBufferWrite!=g.RxBufferRead){ //Zeichen vorhanden
      c = getNextChar();
#if 0
      if ('$'==c){
        clearLcd();
      }else{
        lcdWrite(c);
      }

#else

      switch(status){
        //$GPRMC,161229.487,A,3723.2475,N,12158.3416,W,0.13,309.62,120598, ,*10
        //Parsen: Zeile muss mit $GPRMC beginnen. Speed steht zwischen 7. und 8. Komma. (in Knoten)	    
        case 0: 
            value=0;
            n=0;
            mult=0.1;
            if ('$'==c) status++; break;
        case 1: if ('G'==c) status++; else status=99; break;
        case 2: if ('P'==c) status++; else status=99; break;
        case 3: if ('R'==c) status++; else status=99; break;
        case 4: if ('M'==c) status++; else status=99; break;
        case 5: if ('C'==c) status++; else status=99; break;
        case 6: if (','==c) status++; else status=99; break;
#if 0
        case 7:
          clearLcd();
          status++;
        case 8:
             lcdWrite(c);
             if (','==c) n++; 
            if (5==n){
               status++; 
               home2Lcd();
            } 
            break;
        case 9:
             lcdWrite(c);
             if (','==c) n++; 
             if (9==n) status=99;  
             break;


#else
        case 7: hh=(c-48); status++; break;
        case 8: h=(c-48); status++; break;
        case 9: mm=(c-48); status++; break;
        case 10: m=(c-48); status++; break;
        case 11: ss=(c-48); status++; break;
        case 12: s=(c-48); status++; break;
        case 13: status++; break; //dot
        case 14: z=(c-48); status=20; break; //zehntelsekunden
        case 20: //Komma abzhlen
            if (','==c) n++; 
            if (6==n) status++;  
            break;
        case 21:  //Ganzzahlanteil holen
            if ('.'==c){
                status++;     
            }else{
                if (','==c){
                    status=30;
                }else{
                    value = value * 10.0 + (c-48);	
                }
            }
            break;
        case 22: //Nachkommastellen
            if (','==c){
              status=30;
            }else{
              value += mult * (c-48);
              mult /= 10;
            }
            break;
        case 30: //Zahl auswerten
           if (invalidNMEA){
              invalidNMEA--;
            }
            kmh= value * 1.851984;
            if (kmh >= 0.5){
              LICHT_AN;
            }else{
              LICHT_AUS;
            } 
            x = (U16)(kmh * 100);
            home1Lcd();
            lcdWrite9(hh);
            lcdWrite9(h);
            lcdWrite(':');
            lcdWrite9(mm);
            lcdWrite9(m);
            lcdWrite(':');
            lcdWrite9(ss);
            lcdWrite9(s);
            lcdWrite(',');
            lcdWrite9(z);
            lcdWrite(' ');
            lcdWrite('U');
            lcdWrite('h');
            lcdWrite('r');
            lcdWrite(' ');
            lcdWrite('U');
            lcdWrite('T');
            lcdWrite('C');
            home2Lcd();

            lcdWrite9(x/10000);
            x%=10000;
            lcdWrite9(x/1000);
            x%=1000;
            lcdWrite9(x/100);
            x%=100;
            lcdWrite(',');
            lcdWrite9(x/10);
            lcdWrite9(x%10);
            lcdWrite('K');
            lcdWrite('m');
            lcdWrite('/');
            lcdWrite('h');
            lcdWrite(' ');
            status=99;
            break;
#endif        
        case 99: if (0x0d==c) status=0; break;
        default:
          status=0; break;
      }//switch
#endif
    }//Zeichen vorhanden
  }//while(1)
}
//EOF
